
/*\
  /     Amfow_imi - fido-over-email mailer (c) Alex Kocharin, 2:50/13
 /      
/       This program is free software, licence is GNU GPL
\*/


inline void ReplaceChar(std::string& value, char w, char t)
{
	int s=value.length();
	for(int i=0;i<=s;i++)
		if (value[i] == w)
			value[i] = t;
}

struct string_pair
{
	std::string first;
	std::string second;
};

struct int_pair
{
	int first;
	int second;
};

class C_BSO_Outbound
{
	private:
		regex_t bso_regexp_wp;
		regex_t bso_regexp;
		std::string path;
		void rescan_for_points(std::string indir, std::list<std::string>& result);
		void rescan_for_nodes(std::string indir, std::list<std::string>& result);
		void get_flos_to(std::string link, std::list<std::string>& result);
		void get_plain_to(std::string link, std::list<std::string>& result);
	public:
		C_BSO_Outbound();
		~C_BSO_Outbound();
		void Init(std::string path);
		void GetLinks(std::list<std::string>& result);
		string_pair get_link_filename(std::string nodeaddr);
		void GetMailTo(std::string link,std::list<std::string>& result);
		void DeleteSnt(std::string link,std::string fname);
		void DelFromFlo(std::string floname, std::string snt_file);
		void Process_flo(std::string floname, std::list<std::string>& result);
} Outbound;

int todec(std::string value);
std::string tohex(int value,int chars);

C_BSO_Outbound::C_BSO_Outbound()
{
}

void C_BSO_Outbound::Init(std::string path)
{
	this->path = path;
	regcomp(&this->bso_regexp_wp,"^[0-9a-f]\\{8\\}\\.\\(\\([i,c,f,h,d]lo\\)\\|\\([i,c,o,h,d]ut\\)\\|\\(req\\)\\|\\(pnt\\)\\)$",REG_ICASE | REG_NOSUB);
	regcomp(&this->bso_regexp,"^[0-9a-f]\\{8\\}\\.\\(\\([i,c,f,h,d]lo\\)\\|\\([i,c,o,h,d]ut\\)\\|\\(req\\)\\)$",REG_ICASE | REG_NOSUB);
}

C_BSO_Outbound::~C_BSO_Outbound()
{
	regfree(&this->bso_regexp_wp);
	regfree(&this->bso_regexp);
}

void C_BSO_Outbound::GetLinks(std::list<std::string>& result)
{
	std::string outpath = this->path;
	tolog('n',"Rescanning outbound");
	ReplaceChar(outpath,'\\','/');
	if (outpath.find('/') == std::string::npos) { outpath = "./" + outpath; }
	int p = outpath.rfind('/');
	std::string updirname = outpath.substr(0,p);
	std::string outdirname = outpath.substr(p+1,std::string::npos);
	std::string outdirlc = strtolower(outdirname);
	Directory dir(updirname.c_str());
	std::string file = "";
	while ((file = dir.read()) != "") {
		if (outdirlc == file || (outdirlc+'.') == (file.substr(0,file.length()-3))) { 
			unsigned int pos = file.find('.');
			std::string zone;
			if (pos != std::string::npos) {
				std::string at = file.substr(pos+1,std::string::npos);
				zone = my_itoa(todec(at));
			} else {
				zone = CONFIG.GetVar("address");
				zone = zone.substr(0,zone.find(':'));
			}
			
			std::list<std::string> q;
			this->rescan_for_nodes(updirname+'/'+file,q);
			while (!q.empty()) {
				bool found = false;
				std::string at = zone+':'+q.front();
				std::list<std::string>::const_iterator i;
				for (i = result.begin(); i!=result.end(); i++) {
					if (*i == at) { found = true; break; }
				}
				if (!found) {
					result.push_back(at);
					tolog('n',at);
				}
				q.pop_front();
			}
		}
	}
}

void C_BSO_Outbound::rescan_for_nodes(std::string indir, std::list<std::string>& result)
{
	Directory dir(indir.c_str());
	std::string file = "";
	while ((file = dir.read()) != "") {
		if (regexec(&this->bso_regexp_wp, file.c_str(), 0, 0, 0)==0)
		{
			std::string nn = my_itoa(todec(file.substr(0,4))) + '/' + my_itoa(todec(file.substr(4,4)));
			std::string f = indir+'/'+file;
			if (strtolower(f.substr(f.length()-4,std::string::npos))==".pnt") {
				std::list<std::string> arr;
				rescan_for_points(f, arr);
				while (!arr.empty()) {
					result.push_back(nn+'.'+arr.front());
					arr.pop_front();
				}
			} else {
				result.push_back(nn);
			}
		}
	}
}

void C_BSO_Outbound::rescan_for_points(std::string indir, std::list<std::string>& result)
{
	Directory dir(indir.c_str());
	std::string file = "";
	while ((file = dir.read()) != "") {
		if (regexec(&this->bso_regexp, file.c_str(), 0, 0, 0)==0)
		{
			result.push_back(my_itoa(todec(file.substr(0,8))));
		}
	}
}

string_pair C_BSO_Outbound::get_link_filename(std::string nodeaddr)
{
	std::string path = this->path;
	C_FidoAddr addr = GetFidoAddr(nodeaddr);
	std::string myaddr = CONFIG.GetVar("address");
	if (my_itoa(addr.zone) != myaddr.substr(0,myaddr.find(':'))) {
		path += '.'+tohex(addr.zone,3);
	}
	path += '/';
	std::string file = tohex(addr.net,4) + tohex(addr.node,4);
	if (addr.point) {
		path += file + ".pnt/";
		file = tohex(addr.point,8);
	}
	string_pair result;
	result.first = path;
	result.second = file;
	return result;
}

void C_BSO_Outbound::GetMailTo(std::string link,std::list<std::string>& result)
{
	this->get_plain_to(link, result);
	std::list<std::string> flos;
	this->get_flos_to(link, flos);
	while(!flos.empty()) {
		this->Process_flo(flos.front(), result);
		flos.pop_front();
	}
	std::list<std::string>::const_iterator i;
	bool found = false;
	for(i=result.begin();i!=result.end();i++) {
		found = true;
		tolog('o',"Found file "+(*i)+" for "+link);
	}
	if (!found) tolog('o',"No files found for node "+link);
}

void C_BSO_Outbound::get_flos_to(std::string link, std::list<std::string>& result)
{
	string_pair pf = this->get_link_filename(link);
	Directory dir(pf.first);
	std::string file = "";
	while ((file = dir.read()) != "") {
		if (file.substr(0,8) == pf.second && strtolower(file.substr(10,2)) == "lo") {
			result.push_back(pf.first+'/'+file);
		}
	}
}

void C_BSO_Outbound::get_plain_to(std::string link, std::list<std::string>& result)
{
	string_pair pf = this->get_link_filename(link);
	Directory dir(pf.first);
	std::string file = "";
	while ((file = dir.read()) != "") {
		if (file.substr(0,8) == pf.second && (strtolower(file.substr(10,2)) == "ut" || strtolower(file.substr(9,3)) == "req")) {
			result.push_back(pf.first+'/'+file);
		}
	}
}

void C_BSO_Outbound::DeleteSnt(std::string link,std::string fname)
{
	std::list<std::string> q;
	std::string fnamelc = strtolower(fname);
	this->get_plain_to(link,q);
	while (!q.empty()) {
		if (fnamelc == strtolower(basename(q.front()))) {
			tolog('x',"Xeped");
			unlink(q.front().c_str());
		}
		q.pop_front();
	}
	this->get_flos_to(link,q);
	while (!q.empty())
	{
		this->DelFromFlo(q.front(), fname);
		q.pop_front();
	}
}

void C_BSO_Outbound::DelFromFlo(std::string floname, std::string snt_file)
{
	FILE* file;
	char buf[10240];
	char* result;
	file = fopen(floname.c_str(),"r");
	std::queue<std::string> flo_files;
	std::string sntlo = strtolower(snt_file);
	while((result = fgets(buf,1024,file)) != NULL) {
		std::string line = trim(result);
		if (line != "") {
			char fs = line[0];
			bool fir_upr = false;
			if (fs == '^' || fs == '#' || fs == '~' || fs == '!') {
				fir_upr = true;
				line.erase(0,1);
			}
			if (sntlo == basename(line)) {
				char ss = line[0];
				if (ss != '/' && ss != '\\') {
					unsigned int f = floname.rfind('/');
					if (f != std::string::npos)
						line = floname.substr(0,f-1)+line;
				}
				if (fs == '#') {
					tolog('d',"Cutting file " + line);
					FILE* f = fopen(line.c_str(),"w");
					fclose(f);
				} else if (fs == '^') {
					tolog('d',"Deleting file " + line);
					unlink(line.c_str());
				}
			} else {
				if (fir_upr) line = fs + line;
				flo_files.push(line);
			}
		}
	}
	fclose(file);
	
	if (!flo_files.empty()) {
		file = fopen(floname.c_str(),"w");
		do {
			std::string line = flo_files.front();
			fwrite(line.c_str(),line.length(),1,file);
			fwrite("\n",1,1,file);
			flo_files.pop();
		} while (!flo_files.empty());
		fclose(file);
	} else {
		tolog('d',"Flow "+floname+" is empty - deleting it");
		unlink(floname.c_str());
	}
}

void C_BSO_Outbound::Process_flo(std::string floname, std::list<std::string>& result)
{
	FILE* file = fopen(floname.c_str(),"r");
	char buf[10240];
	char* res;
	while((res = fgets(buf,1024,file)) != NULL) {
		std::string line = trim(res);
		if (line != "") {
			char fs = line[0];
			if (fs == '^' || fs == '#' || fs == '!' || fs == ' ' || fs == '~'){
				line.erase(0,1);
			}
			char ss = line[0];
			if ((ss != '/') && (ss != '\\') && (line.find(':')==std::string::npos)) { 
				unsigned int f = floname.rfind('/');
				if (f != std::string::npos)
					line = floname.substr(0,f)+line;
			}
			if (line != "" && fs != '~' && fs != '!') {
				FILE* test = fopen(line.c_str(),"r");
				if (test != NULL) {
					result.push_back(line);
					fclose(test);
				}
			}
		}
	}
	fclose(file);
}

//}

/*
<?

function A_StartSession($address = '', $client = 0, $is_d = 0)
{
	GLOBAL $InData;
	GLOBAL $OutData;
	GLOBAL $CONFIG;
	GLOBAL $STime;
	$STime = mktime();
	$OutData = new C_Out_Data;
	$OutData->is_data = $is_d;
	$OutData->SendEMSI();
	if ($client):
		$have_pwd = (strcmp(@$CONFIG->Links->Vals[$address]['password'],'')!=0);
		if ($have_pwd):
		$OutData->SetCrypt($CONFIG->Links->Vals[$address]['password']);
		endif;
		$InData->Stage_Init($have_pwd?2:1);
		tolog('~',"~P ".($have_pwd?'pwd':'none'));
	else:
		$InData->Stage_Init();
		if (($InData->Sec == 2) && ($pad = $InData->PwdAddress)):
		$OutData->SetCrypt($CONFIG->Links->Vals[$pad]['password']);
		endif;
		$s = $InData->Sec;
		tolog('~',mktime()."~S amfow_server ".((isset($pad) && $pad)?$pad:$InData->Addresses[0]));
		tolog('~',"~P ".($s==0?'err':($s==1?'none':'pwd')));
	endif;
	if ($InData->Error):
	tolog('E','Error: '.$InData->Error);
	else:
	$InData->Stage_Data();
	endif;
	$InData->End_Data();
	$OutData->Write('Q');
	if (!$client) tolog('~',mktime().'~D');
}

function TraceData($file)
{
	GLOBAL $DATA;
	$data = 0;
	$cnt = 0;
	if ($file):
		while (!feof($file)):
			$d = fread($file,10);
//			tolog('2',$d);
			$DATA.=$d;
//			if (strstr($d1.$d,'**DATA(')) { tolog('5',$d1.$d); };
			$d1 = $d;
		endwhile;
	endif;
	if ($str = strstr($DATA,'**DATA(')):
		$DATA = substr($str,7);
		list($len,$DATA) = split(')',$DATA);
		$DATA = substr($DATA,1,$len);
	else:
		$DATA = '';
	endif;
}

class C_In_DATA
{
	var $Addresses;
	var $Password;
	var $Error;
	var $block;
	var $FILE;
	var $Sec;
	var $DATA;
	var $RemTime;
	var $PwdAddress;
	var $Rem_Opt;
	var $base64;
	var $Got;
	var $ansfile;
	var $ansto;

	function ChkSec()
	{
		if ($this->Error):
			$this->Sec = 0;
		endif;
	}

	function FromInteger($int)
	{
	$result=0;
	$len = strlen($int);
	for ($i=0;$i<$len;$i++):
		$result+=(ord(substr($int,$i,1)) << (($len-$i-1)*8));
	endfor;
	return $result;
	}

	function ProcAddress()
	{
		GLOBAL $CONFIG;
		$Addresses = split(' ',$this->block);
		for ($i=0;$i<sizeof($Addresses);$i++):
			$Addresses[$i] = strtok($Addresses[$i],'@');
		endfor;
		$this->Sec = 1;
		foreach($Addresses as $addr):
			tolog('A',"Remote AKA: $addr");
			$this->Addresses[] = $addr;
			if (strcmp($this->Password,'')==0):
				if ($pwd = $CONFIG->GetLinkVar($addr,'password')):
				$this->Password = $pwd;
				$this->PwdAddress = $addr;
				$this->Sec = 0;
				endif;
			endif;
		endforeach;
	}

	function ProcHuman()
	{
		tolog('H','   '.$this->block);
	}

	function ProcOpt()
	{
		$this->Rem_Opt = $this->block;
	}

	function ProcList()
	{
		GLOBAL $OutData;
		if ($this->Sec):
		list($md5,$fsize,$fname) = split(',',$this->block,3);
		tolog('G',"Getting {$this->block}");
		$OutData->is_data = 1;
		$OutData->Write('G',$this->block);
		endif;
	}

	function ProcGet()
	{
		tolog('S','Sending '.$this->block);
		GLOBAL $OutData;
		GLOBAL $Outbound;
		if ($this->Sec):
		list($crc,$fsize,$offset,$fname) = split(',',$this->block);
		$this->Got[] = $fname;
		if ($this->Addresses):
		reset($this->Addresses);
		foreach ($this->Addresses as $addr):
			if ($mail = $Outbound->GetMailTo($addr)):
			foreach ($mail as $file):
				if (strcasecmp($fname,basename($file))==0):
					$OutData->SendFile($file);
					$OutData->is_data = 1;
				endif;
			endforeach;
			endif;
		endforeach;
		endif;
		endif;
	}

	function ProcFAttr()
	{
		list($crc,$size,$offset,$name) = split(',',$this->block);
		$this->File[0] = $name;
		$this->File[1] = $crc;
		$this->File[2] = $size;
		$this->File[3] = $offset;
	}

	function ProcTime()
	{
		$this->RemTime = $this->block;
	}

	function ProcCrypt($data)
	{
		tolog('Y',$md5 = md5($this->RemTime.$this->Password));
		$cnt = 0;
		$result = '';
		$this->Sec = 2;
		for ($i=0; $i<strlen($data); $i++):
		$result.=chr(ord(substr($data,$i,1)) ^ ord(substr($md5,$cnt,1)));
		if ((++$cnt) >= strlen($md5)) 
		 { $cnt = 0; };
		endfor;
		return $result;
	}

	function ProcFile()
	{
	GLOBAL $OutData;
	GLOBAL $CONFIG;
	if ($this->Sec):
		$fname = $this->File[0];
		if (preg_match('/[0-9a-f]{8}\.[odhic]ut/i',$fname)) 
			$fname = rand(10000000,99999999).'.pkt';
		tolog("C","Received file $fname");
		tolog('~',"~R ".$this->File[2]);
		if (SaveFileToInbound($fname,$this->block,$this->Sec==2)):
			$OutData->Write('R',$this->File[1].','.$this->File[2].','.$this->File[3].','.$this->File[0]);
			$OutData->is_data = 1;
		endif;
	endif;
	$this->File = array();
	}

	function ProcRcvd()
	{
		GLOBAL $Outbound;
		GLOBAL $OutData;
		list($crc,$fsize,$offset,$fname) = split(',',$this->block);
		$this->Got[] = $fname;
		foreach($this->Addresses as $addr):
	       	$Outbound->DeleteSnt($addr,$fname);
		endforeach;
		tolog('D',"Sent ".$fname);
		tolog('~',"~T ".$fsize);
		$OutData->is_data = 1;
		$OutData->Write('C',$this->block);
	}

	function ProcRvC()
	{

	}
	
	function C_In_DATA($DATA)
	{
		GLOBAL $CONFIG;
		$s = 20;
		if (($ss=strlen($DATA)-2)<$s) $s = $ss;
		$this->base64 = preg_match('/^[A-Za-z0-9\/\+\=]*$/',substr($DATA,0,$s));
		if ($this->base64):
			tolog('O',"Using base64 encoding");
			$DATA = base64_decode($DATA);
		endif;
		if ($ansto = $CONFIG->GetVar('answertofile')):
			$this->ansfile = fopen($ansto,'a');
			$this->ansto = $ansto;
		endif;
		$this->DATA = strstr($DATA,'!');
	}

	function Stage_Init($seclvl = -1)
	{
		tolog('z','Init stage started');
		$DATA = $this->DATA;
		while ($DATA || strlen($DATA)):
		if (strcmp(substr($DATA,0,1),'!')==0):
		$cmd = substr($DATA,1,1);
		if (in_array($cmd,array('A','T','H','Y','O'))):
			$size = substr($DATA,2,2);
			$size_n = $this->FromInteger($size);
			$this->block = substr($DATA,4,$size_n);
			$DATA = substr($DATA,4+$size_n);
			if (isset($this->ansfile)) fwrite($this->ansfile,'!'.$cmd.$size.$this->block);
		switch ($cmd):
			case ('A'):
				$this->ProcAddress();
				break;
			case ('Y'):
				$DATA = $this->ProcCrypt($DATA);
				break;
			case ('T'):
				$this->ProcTime();
				break;
			case ('H'):
				$this->ProcHuman();
				break;
			case ('O'):
				$this->ProcOpt();
				break;
		endswitch;
		else:
			break;
		endif;
		else:
			$this->Error = 'Bad request';
			break;
		endif;
		endwhile;
		if ($seclvl != -1):
			$this->Sec = $seclvl;
		else:
			$this->ChkSec();
		endif;
		tolog('P',"Set ".(int)$this->Sec." secure level");
		$this->DATA = $DATA;
	}

	function Stage_Data()
	{
		GLOBAL $OutData;
		$quit = 0;
		tolog('z','Data stage started');
		$DATA = $this->DATA;
		while ($DATA || strlen($DATA) || !$quit):
		if ((strcmp(substr($DATA,0,1),'!')==0) && (in_array(substr($DATA,1,1),array('E','G','R','C','L','S','F','Q')))):
		$cmd = substr($DATA,1,1);
			if (strcmp($cmd,'F')==0):
				$size = substr($DATA,2,4);
				$size_n = $this->FromInteger($size);
				$this->block = substr($DATA,6,$size_n);
				$DATA = substr($DATA,6+$size_n);
				if (isset($this->ansfile)) { fwrite($this->ansfile,'!'.$cmd.$size.$this->block); };
			else:
				$size = substr($DATA,2,2);
				$size_n = $this->FromInteger($size);
				$this->block = substr($DATA,4,$size_n);
				$DATA = substr($DATA,4+$size_n);
				if (isset($this->ansfile)) fwrite($this->ansfile,'!'.$cmd.$size.$this->block);
			endif;
			switch ($cmd):
			case ('L'):
				$this->ProcList();
				break;
			case ('S'):
				$this->ProcFAttr();
				break;
			case ('F'):
				$this->ProcFile();
				break;
			case ('R'):
				$this->ProcRcvd();
				break;
			case ('G'):
				$this->ProcGet();
				break;
			case ('C'):
				$this->ProcRvC();
				break;
			case ('E'):
				$quit = 1;
				break;
			case ('Q'):
				$quit = 1;
				break;
			endswitch;
		else:
			$this->Error = 'Bad request';
			break;
		endif;
		endwhile;
		if (!$quit) { $OutData->is_data = 1; }
		if ($this->Sec):
		$this->SendList();
		endif;
	}

	function End_Data()
	{
		if (isset($this->ansfile)):
	   		fwrite($this->ansfile,"\n\n");
			fclose($this->ansfile);
			@chmod($this->ansto,0666);
		endif;
	}
	
	function SendList()
	{
		GLOBAL $OutData;
		GLOBAL $Outbound;
		if ($this->Addresses):
		reset($this->Addresses);
		foreach ($this->Addresses as $addr):
			if ($mail = $Outbound->GetMailTo($addr)):
			foreach ($mail as $file):
				$list = true;
				if ($this->Got):
				$bs = basename($file);
					foreach($this->Got as $got):
						if (strcmp($bs,$got)==0):
							$list = false;
							break;
						endif;
					endforeach;
				endif;
				if ($list):
					$OutData->ListFile($file);
					$OutData->is_data = 1;
				endif;
			endforeach;
			endif;
		endforeach;
		endif;
	}
}
*/
bool SaveFileToInbound(std::string fname, std::string data,int Sec)
{
	std::string dir;
	if (Sec > 0) {
		dir = CONFIG.GetVar("inbound");
	} else {
		dir = CONFIG.GetVar("unsecinbound");
	}
	dir += '/';
	std::string i = "";
	FILE* f;
	f = fopen((dir+fname+i).c_str(),"r");
	while (f != NULL) {
		fclose(f);
		int k = atoi(i.c_str())+1;
		i = my_itoa(k);
		f = fopen((dir+fname+i).c_str(),"r");
	}
	
	std::string fnm;
	f = fopen((fnm = dir+fname+i).c_str(),"w");
	if (f != NULL) {
		fwrite(data.c_str(),data.length(),1,f);
		fclose(f);
		tolog('W',"Saving "+fname+" to "+fnm);
//		@chmod($fnm,0666);
//		foreach($CONFIG->GetArray('execonfile') as $exof):
//			list($preg,$exec) = param_split($exof);
//			$exec = str_replace('@file',$fnm,$exec);
//			if (preg_match($preg,$fname)):
//				tolog('W',"Executing $exec");
//				`$exec`;
//			endif;
//		endforeach;
//		foreach($CONFIG->GetArray('includeonfile') as $exof):
//			list($preg,$exec) = param_split($exof);
//			$exec = str_replace('@file',$fnm,$exec);
//			if (preg_match($preg,$fname)):
//				tolog('W',"Including $exec");
//				include $exec;
//			endif;
//		endforeach;
		return true;
	} else {
		tolog('E',"Error writing "+fnm);
		return false;
	}
}
/*
class C_Out_DATA
{
	var $DATA;
	var $is_data;
	var $Y_md5;
	var $Y_cnt;

	function C_Out_Data()
	{
		$this->Y_md5 = '';
	}
	
	function SendEMSI()
	{
	GLOBAL $CONFIG;
	GLOBAL $Prot_session;
	GLOBAL $STime;
	$addrstr = '';
	$this->Write('A',join(' ',$CONFIG->GetArray('address')));
	$this->Write('T',$STime);
	$this->Write('H','System: '.$CONFIG->Vars['system']);
	$this->Write('H','Sysop: '.$CONFIG->Vars['sysop']);
	$this->Write('H','Place: '.$CONFIG->Vars['location']);
	$this->Write('H',"Mailer: Amfow/".VERSION."/".($CONFIG->Vars['servermode']?'Server':'Client'));
#	$this->Write('H',"Time: ".date("r, t"));
	}

	function SetCrypt($pwd)
	{
		GLOBAL $STime;
		$this->Write('Y');
		$this->Y_md5 = md5($STime.$pwd);
		tolog('Y',md5($STime.$pwd));
		$this->Y_cnt = 0;
	}

	function ListFile($filename,$as_file='')
	{
		if (strcmp($as_file,'')==0) { $as_file = basename($filename); }
		$File_data = file_get_contents($filename);
		$this->Write('L',GetCRC($File_data).','.filesize($filename).',0,'.$as_file);
	}

	function SendFile($filename,$as_file='')
	{
		if (strcmp($as_file,'')==0) { $as_file = basename($filename); }
//		$this->Write('(','');
//		$this->Write('N',$as_file);
		$File_data = file_get_contents($filename);
		$this->Write('S',GetCRC($File_data).','.filesize($filename).',0,'.$as_file);
		$this->Write('F',$File_data,4);
//		$this->Write(')','');
	}

	function ToInteger($data,$len)
	{
	$result = '';
	for ($i=$len-1;$i>=0;$i--):
		$result.=chr(($data >> ($i*8)) & 255);
	endfor;
	return $result;
	}

	function Write($param,$value = '',$len = 2)
	{
	$adata = '';
	$newdata = '!'.$param.$this->ToInteger(strlen($value),$len).$value;
	if ($this->Y_md5):
		for ($i=0; $i<strlen($newdata); $i++):
		$adata.=chr(ord(substr($newdata,$i,1)) ^ ord(substr($this->Y_md5,$this->Y_cnt,1)));
		if ((++$this->Y_cnt) >= strlen($this->Y_md5)) 
		 { $this->Y_cnt = 0; };
		endfor;
	else:
		$adata = $newdata;
	endif;
	$this->DATA.=$adata;
	}
}
*/

std::string tohex(int value,int chars)
{
	std::string result = "";
	std::string msv = "0123456789abcdef";
	for (int i = chars-1;i>=0;i--)
	{
		result += msv[value/(1 << (i*4))];
		value%=1<<(i*4);
	}
	return result;
}

int todec(std::string value)
{
	int result=0;
	value = strtolower(value);
	int len = value.length();
	for (int i=0;i<len;i++)
	{
		int v = int(value[i])-int('0');
		if (v>9) { v=int(value[i])-int('a')+10; }
		result+=(v<<((len-i-1)*4));
	}
	return result;
}

