
/*\
  /     Amfow_imi - fido-over-email mailer (c) Alex Kocharin, 2:50/13
 /      
/       This program is free software, licence is GNU GPL
\*/


struct smtp_record
{
	std::string host;
	std::string port;
	std::string user;
	std::string pass;
};

void Receive_mail_from(std::string popsrv);
bool Process_message_Aemail(std::string message);
void Send_mail_via(smtp_record& s, std::queue<std::string> links);
std::string File_to_email_msg(std::string file, std::string as_file, std::string from, std::string to, std::string subj);


void Receive_mail()
{
	std::queue<std::string> q;
	CONFIG.GetArray("pop3server",q);
	while (!q.empty()) {
		Receive_mail_from(q.front());
		q.pop();
	}
}

void Receive_mail_from(std::string popsrv)
{
	bool error = false;
	std::string id = "", hostport = "", login = "", tmp = "", host = "", port = "", user = "", pass = "";
	split_2(' ', popsrv, id, tmp);
	split_2(' ', tmp, hostport, login);
	tolog('1',"Receiving mail from " + id);
	split_2(':', hostport, host, port);
	if (port == "") port = "110";
	tolog('1',"Connecting with "+host+':'+port);
	split_2(':', login, user, pass);
	C_pop3 pop;
	if (pop.Connect(host, port)) {
		if (pop.Login(user,pass)) {
			string_pair stats = pop.Stat();
			for (int i=1;i<=atoi(stats.first.c_str());i++) {
				if (!Process_message_Aemail(pop.GetMessage(i))) {
					error = 1;
					break;
				}
			}
			pop.Quit();
			if (error) {
				tolog('1',"Error during the session");
			} else {
				tolog('1',"Session completed");
			}
		} else {
			pop.Quit();
			tolog('1',"Access denied");
		}
	} else {
		tolog('1',"Cannot connect to "+host+':'+port);
	}
}

bool Process_message_Aemail(std::string message)
{
	std::string tempin;
	if ((tempin = CONFIG.GetVar("tempinbound"))!="") {
		std::string i = "";
		FILE* f;
		f = fopen((tempin+"/msg"+i+".ema").c_str(),"r");
		while (f != NULL) {
			fclose(f);
			int k = atoi(i.c_str())+1;
			i = my_itoa(k);
			f = fopen((tempin+"/msg"+i+".ema").c_str(),"r");
		}
		std::string fname = tempin+"/msg"+i+".ema";
		if ((f = fopen(fname.c_str(),"w")) != NULL) {
			fwrite(message.c_str(),message.length(),1,f);
			fclose(f);
			return true;
		} else {
			tolog('1',"Error opening file "+fname);
			return false;
		}
	} else {
		tolog('1',"Error: tempinbound not defined in config");
		return false;
	}
}

void Send_mail()
{
	std::map< std::string,std::queue<std::string> > polls;
	std::list<std::string> nodes;
	Outbound.GetLinks(nodes);
	if (!nodes.empty()) {
		do {
			std::string link = nodes.front();
			nodes.pop_front();
			std::list<std::string> mail;
			Outbound.GetMailTo(link, mail);
			if (!mail.empty()) {
				std::string emailid = CONFIG.GetLinkVar(link,"sendvia");
				if (emailid != "") {
					polls[emailid].push(link);
				} else {
					tolog('l',link+" is not email link");
				}
			}
		} while (!nodes.empty());
	} else {
		tolog('l',"No outgoing mail for EON's");
	}

	if (!polls.empty()) {
		std::map<std::string,smtp_record> smtps;
		std::queue<std::string> q;
		CONFIG.GetArray("smtpserver",q);
		while (!q.empty()) {
			std::string id = "", tmp1 = "", tmp2 = "", host_port = "", user_pass = "";
			split_2(' ',q.front(),id,tmp1);
			q.pop();
			split_2(' ',tmp1,host_port,user_pass);
			smtp_record smtp;
			split_2(':',host_port,tmp1,tmp2);
			smtp.host = tmp1;
			smtp.port = tmp2;
			split_2(':',user_pass,tmp1,tmp2);
			smtp.user = tmp1;
			smtp.pass = tmp2;
			smtps[id] = smtp;
		}

		std::map< std::string,std::queue<std::string> >::iterator it;
		for(it = polls.begin(); it != polls.end(); it++) {
			Send_mail_via(smtps[it->first],it->second);
		}
	}
}

void Send_mail_via(smtp_record& s, std::queue<std::string> links)
{
	if (s.port == "") s.port = "25";
	tolog('C',"Connecting to "+s.host+':'+s.port);
	C_smtp smtp;
	bool error = false;
	if (smtp.Connect(s.host,s.port)) {
		if (s.user == "-" || smtp.Login(s.user,s.pass)) {
			while (!links.empty() && !error) {
				std::string link = links.front();
				links.pop();
				tolog('S',"Send mail for "+link);
				tolog('~',"~P pwd");
				tolog('~',my_itoa(time(0))+"~S amfow_imi "+link);
				std::list<std::string> mail;
				Outbound.GetMailTo(link, mail);
				while (!mail.empty()) {
					std::string file = mail.front();
					mail.pop_front();
					std::string from = CONFIG.GetLinkVar(link,"sendfrom");
					std::string to = CONFIG.GetLinkVar(link,"sendto");
					std::string subj = CONFIG.GetLinkVar(link,"sendsubj");
					std::string fbase = basename(file);
					std::string as_file = rand_if_pkt(fbase);
					std::string message = File_to_email_msg(file,as_file,from,to,subj);
					tolog('~',"~T "+my_itoa(message.length()));
					if (!smtp.Helo(s.host)) { error = true; break; };
					if (!smtp.Rset()) { error = true; break; };
					if (!smtp.SendMessage(message,from,to)) { error = true; break; };
					tolog('f',"Sent file "+file+" as "+as_file);
					Outbound.DeleteSnt(link,fbase);
				}
				tolog('~',my_itoa(time(0))+"~D");
			}
			if (!smtp.Rset()) error = 1;
			smtp.Quit();
			if (error) {
				tolog('1',"Error during the session");
			} else {
				tolog('1',"Session completed");
			}
		} else {
			smtp.Quit();
			tolog('1',"Access denied");
		}
	} else {
		tolog('1',"Cannot connect to "+s.host+':'+s.port);
	}
}

std::string File_to_email_msg(std::string file, std::string as_file, std::string from, std::string to, std::string subj)
{
	std::string boundid = "==_" + my_itoa((10000000+((int)(((double)rand()/RAND_MAX)*89999999)))) + "_" + my_itoa((10000000+((int)(((double)rand()/RAND_MAX)*89999999)))) + "_==";
	std::string result = "Mailer: Amfow_imi/";
	result+= VERSION;
	result+= "\r\n";
	result+= "To: "+to+"\r\n";
	result+= "From: "+from+"\r\n";
	result+= "Subject: "+subj+"\r\n";
	result+= "Content-Type: multipart/mixed; boundary=\""+boundid+"\"\r\n";
	result+= "\r\n";
	result+= "--"+boundid+"\r\n";
	result+= "Content-Type: application/octet-stream; name=\""+as_file+"\"\r\n";
	result+= "Content-Disposition: attachment; filename=\""+as_file+"\"\r\n";
	result+= "Content-Transfer-Encoding: base64\r\n";
	result+= "\r\n";
	std::string base64 = base64_encode(file_get_contents(file));
	std::string data;
	while (base64.length() > 60) {
		data = base64.substr(0,60);
		base64.erase(0,60);
		result+= data+"\r\n";
	}
	result+= base64+"\r\n\r\r";
	result+= "--"+boundid;
	return result;
}

int ExtractFile(FILE* f, std::string fname,int sec)
{
	std::string b64 = "";
	char buf[10240];
	char* result;
	while((result = fgets(buf,1024,f)) != NULL) {
		std::string line = trim(result);
		if (line.length() == 0) break;
	}
	while((result = fgets(buf,1024,f)) != NULL) {
		std::string line = trim(result);
		if (line.substr(0,2) == "--") {
			break;
		} else {
			b64 += line;
		}
	}
	if (fname != "") {
		std::string file = base64_decode(b64);
		if (!SaveFileToInbound(fname, file, sec))
			tolog('E',"Could not save file "+fname);
		return file.length();
	} else {
		return 0;
	}
}

std::string ChkFields(std::string from, std::string to, std::string subj)
{
	std::map< std::string,std::map<std::string,std::string> >::const_iterator i;
	for (i=CONFIG.Links.Vals.begin(); i!=CONFIG.Links.Vals.end(); i++) {
		std::string aka = i->first;
		std::map<std::string,std::string>::const_iterator mapi;
		std::string fromf = "", tof = "", subjf = "";
		if ((mapi = i->second.find("fieldfrom")) != i->second.end())
			fromf = mapi->second;
		if ((mapi = i->second.find("fieldto")) != i->second.end())
			tof = mapi->second;
		if ((mapi = i->second.find("fieldsubj")) != i->second.end())
			subjf = mapi->second;
		if (fromf != "" || tof != "" || subjf != "") {
			if ((fromf == "" || (from.find(fromf) != std::string::npos)) &&
				(tof == "" || (to.find(tof) != std::string::npos)) &&
				(subjf == "" || (subj.find(subjf) != std::string::npos))) {
				if ((mapi = i->second.find("sysop")) != i->second.end())
					tolog('1',"Bundle from "+mapi->second);
				return aka;
			}
		}
	}
	tolog('1',"Unsecure bundle");
	return "";
}

void ExtractImi(std::string fname)
{
	FILE* f = fopen(fname.c_str(),"r");
	int sec = 0;
	std::string fn = "", subj = "", from = "", to = "", aka = "";
	char buf[10240];
	char* result;
	while((result = fgets(buf,1024,f)) != NULL) {
		std::string line = trim(result);
		std::string str = "", val = "";
		split_2(':',line,str,val);
		std::string strlo = strtolower(str);
		if (subj == "" && strlo == "subject") {
			subj = val;
		} else if (from == "" && strlo == "from") {
			from = val;
		} else if (to == "" && strlo == "to") {
			to = val;
		} else {
			std::string linelo = strtolower(line);
			if (linelo.substr(0,13) == "content-type:") {
				unsigned int ps = linelo.find("name=");
				if (ps != std::string::npos) {
					std::string fn = line.substr(ps+5, std::string::npos);
					fn = fn.substr(0,fn.find(' '));
					if (fn[0]=='"') {
						fn.erase(0,1);
						fn = fn.substr(0,fn.find('"'));
					}
					if ((aka = ChkFields(from,to,subj))!="") {
						sec = 1; 
					} else {
						sec = 0;
					}
					ExtractFile(f,fn,sec);
				}
			}
		}
	}
	fclose(f);
	tolog('~',my_itoa(time(0))+"~S amfow_imi "+aka);
	if (aka != "") {
		tolog('~',"~P pwd");
	} else {
		tolog('~',"~P none");
	}
	tolog('~',"~R " + filesize(fname));
	tolog('~',my_itoa(time(0))+"~D");
	if (sec == 1) {
		unlink(fname.c_str()); 
	} else {
		std::string bcp = CONFIG.GetVar("bcpdir");
		if (bcp != "") { 
			std::string i = "";
			FILE* f;
			f = fopen((bcp+"/msg"+i+".ema").c_str(),"r");
			while (f != NULL) {
				fclose(f);
				int k = atoi(i.c_str())+1;
				i = my_itoa(k);
				f = fopen((bcp+"/msg"+i+".ema").c_str(),"r");
			}
			rename(fname.c_str(),(bcp+"/msg"+i+".ema").c_str()); 
		} 
		else { 
			unlink(fname.c_str());
		}
	}
}

void ExtractFiles()
{
	std::string tmpin = CONFIG.GetVar("tempinbound");
	Directory dir(tmpin);
	std::string file;
	while ((file = dir.read()) != "") {
		unsigned int ps = file.find('.');
		if (ps != std::string::npos) {
			if (file.substr(0,3) == "msg") {
				if (file.substr(ps,std::string::npos) == ".ema") {
					bool is_ema = true;
					for (unsigned int i=3;i<ps;i++) {
						int ord = (int)file[i];
						if (!(ord >= (int)'0' && ord <= (int)'9')) {
							is_ema = false;
							break;
						}
					}
					if (is_ema) {
						ExtractImi(tmpin+'/'+file);
					}
				}
			}
		}
	}
}

