
/*\
  /     Amfow_imi - fido-over-email mailer (c) Alex Kocharin, 2:50/13
 /      
/       This program is free software, licence is GNU GPL
\*/


class C_pop3
{
	private:
		std::string request;
		std::string answer;
		C_network sock;
		void _disconnect();
		bool _sendstr(std::string str = "");
	public:
		bool Connect(std::string host, std::string port);
		bool Login(std::string user, std::string pass);
		string_pair Stat();
		std::string GetMessage(int msg);
		void Quit();
};

class C_smtp
{
	private:
		std::string request;
		std::string answer;
		C_network sock;
		void _disconnect();
		int _sendstr(std::string str = "");
	public:
		bool Connect(std::string host, std::string port);
		bool Login(std::string user, std::string pass);
		bool Rset();
		bool Helo(std::string host);
		string_pair Stat();
		bool SendMessage(std::string msg, std::string from, std::string to);
		void Quit();
};

bool C_pop3::Connect(std::string host, std::string port)
{
	if (this->sock.Connect(host, atoi(port.c_str()))) {
		if (!this->_sendstr()) {
			this->sock.Disconnect();
			return false;
		} else {
			return true;
		}
	}
	return false;
}

bool C_pop3::Login(std::string user, std::string pass)
{
	if (!this->_sendstr("USER "+user)) return false;
	if (!this->_sendstr("PASS "+pass)) return false;
	return true;
}

string_pair C_pop3::Stat()
{
	string_pair result;
	result.first = result.second = "";
	if (this->_sendstr("STAT")) {
		std::string ans = "", num = "", sz = "", tmp = "";
		split_2(' ',this->answer,ans,tmp);
		split_2(' ', tmp, num, sz);
		result.first = num;
		result.second = sz;
	}
	return result;
}

void C_pop3::_disconnect()
{
	this->sock.Disconnect();
}

std::string C_pop3::GetMessage(int msg)
{
	if (this->_sendstr("RETR "+my_itoa(msg))) {
		std::string line = "", result = "";
		while(trim(line) != ".") {
			result += line;
			line = this->sock.reads(10240);
			if (line == "") break;
		}
		this->_sendstr("DELE "+my_itoa(msg));
		return result;
	} else {
		return "";
	}
}

void C_pop3::Quit()
{
	this->_sendstr("QUIT");
	this->_disconnect();
}

bool C_pop3::_sendstr(std::string str)
{
	this->request = str;
	if (str != "") {
		tolog('X',"S: "+str);
		this->sock.write(str+"\r\n");
	}
	std::string line = this->sock.reads(10240);
	this->answer = line;
	tolog('X',"R: " + chop(line));
	return line.substr(0,3) == "+OK";
}

bool C_smtp::Connect(std::string host, std::string port)
{
	if (this->sock.Connect(host, atoi(port.c_str()))) {
		if (this->_sendstr() != 220) {
			this->sock.Disconnect();
			return false;
		} else {
			return true;
		}
	}
	return false;
}

bool C_smtp::Login(std::string user, std::string pass)
{
	if (this->_sendstr("AUTH LOGIN")!=334) return 0;
	if (this->_sendstr(base64_encode(user))!=334) return 0;
	if (this->_sendstr(base64_encode(pass))!=235) return 0;
	return true;
}

bool C_smtp::Rset()
{
	return this->_sendstr("RSET")==250;
}

bool C_smtp::Helo(std::string host)
{
	return this->_sendstr("HELO "+host)==250;
}

string_pair C_smtp::Stat()
{
	string_pair result;
	result.first = result.second = "";
	if (this->_sendstr("STAT")) {
		std::string ans = "", num = "", sz = "", tmp = "";
		split_2(' ',this->answer,ans,tmp);
		split_2(' ', tmp, num, sz);
		result.first = num;
		result.second = sz;
	}
	return result;
}

void C_smtp::_disconnect()
{
	this->sock.Disconnect();
}

bool C_smtp::SendMessage(std::string msg, std::string from, std::string to)
{
	unsigned int k, p = 0;
	while ((k = msg.find('\r')) != std::string::npos) msg.erase(k,1);
	while ((k = msg.find("\n.\n")) != std::string::npos) msg.replace(k,3,"\n. \n");
	while ((k = msg.find("\n",p)) != std::string::npos) {
		msg.replace(k,1,"\r\n");
		p = k + 2;
	}
	if ((this->_sendstr("MAIL FROM:<"+from+">")==250) &&
		(this->_sendstr("RCPT TO:<"+to+">")==250) &&
		(this->_sendstr("DATA")==354)) {
		this->sock.write(msg + "\r\n.\r\n");
		return this->_sendstr()==250;
	} else {
		return 0;
	}
}

void C_smtp::Quit()
{
	this->_sendstr("QUIT");
	this->_disconnect();
}

int C_smtp::_sendstr(std::string str)
{
	this->answer = "";
	this->request = str;
	if (str != "") {
		tolog('X',"S: "+str);
		this->sock.write(str+"\r\n");
	}
	std::string line;
	int ps;
	do {
		line = this->sock.reads(10240);
		this->answer += line;
		tolog('X',"R: "+chop(line));
		ps = line.find(' ');
	} while ((ps>3)||(ps<2));
	return atoi(line.substr(0,line.find(' ')).c_str());
}

