
/*\
  /     Amfow_imi - fido-over-email mailer (c) Alex Kocharin, 2:50/13
 /      
/       This program is free software, licence is GNU GPL
\*/

#define VERSION "0.96"

class C_LOG
{
	private:
		FILE* file;
		std::string lvls;
		std::string br;
		bool is_stdout;
		std::string filename;
	public:
		C_LOG(std::string fname,std::string lvls,std::string br);
		~C_LOG();
		void write(char lvl,std::string message);
		void refseek();
};

std::list<C_LOG*> LOGS;

class C_configfile_reader
{
	private:
		FILE* file;
		int nofile;
		int iflvl;
		std::stack<int> Levels;
		std::stack<int> LvlsPs;
		int is_equal(std::string str);
		void insert_constant(std::string &spec, std::string &chars, int b, int e);
	public:
		int linenum;
		C_configfile_reader(std::string filename = "");
		~C_configfile_reader();
		std::string ReadStr(std::string line = "");
		std::string read_and_translate(std::string line);
};

struct E_end_input
{

};

void aks_init(std::string module = "Unknown")
{
	CONFIG.Open("config",module);
	initlog();
	Outbound.Init(CONFIG.GetVar("outbound"));
	time_t t = time(0);
	srand(t);
}

void aks_done()
{
	donelog();
	GetFidoAddr("",true);
}

void initlog()
{
	std::queue<std::string> q;
	CONFIG.GetArray("logfile",q);
	while (!q.empty()) {
		std::string logfile = q.front();
		q.pop();
		std::queue<std::string> logfarr;
		param_split(logfile, logfarr);
		std::string fname = logfarr.front();
		logfarr.pop();
		std::string lvls = "", br = "\n";
		if (!logfarr.empty()) {
			lvls = logfarr.front();
			logfarr.pop();
			if (!logfarr.empty()) {
				logfarr.pop();
			}
		}
		C_LOG* newlog = new C_LOG(fname,lvls,br);
		LOGS.push_back(newlog);
	}
}

extern void tolog(char lvl,std::string message)
{
	std::list<C_LOG*>::iterator iter;
	for (iter=LOGS.begin(); iter!=LOGS.end(); iter++)
		(*iter)->write(lvl,message);
}

extern void donelog()
{
	std::list<C_LOG*>::iterator iter;
	for (iter=LOGS.begin(); iter!=LOGS.end(); iter++)
		delete (*iter);
	LOGS.clear();
}

extern void refseeklog()
{
	std::list<C_LOG*>::iterator iter;
	for (iter=LOGS.begin(); iter!=LOGS.end(); iter++)
		(*iter)->refseek();
}

void C_Links::OpenCfg(std::string filename)
{
	bool linknew = false;
	std::map<std::string,std::string> Curr_ms;
	std::map<std::string,C_Links_loc> Curr_ms_loc;
	this->_filename = filename;
	C_configfile_reader reader(filename);
	try {
		while (1) {
			std::string line = reader.ReadStr();
//std::cout << line << "\n";
			if (line != "") {
				std::string key, value;
				split_2(' ',line,key,value);
				key = strtolower(key);
				if (key == "linkdefaults") {
					this->AddLink(linknew, Curr_ms, Curr_ms_loc);
					linknew = false;
				} else if (key == "link") {
					this->_line = reader.linenum;
					this->AddLink(linknew, Curr_ms, Curr_ms_loc);
					linknew = true;
					Curr_ms["sysop"] = value;
				} else {
					std::string i = "";
					while (Curr_ms.find(key+i) != Curr_ms.end()) {
						int k = atoi(i.c_str())+1;
						i = my_itoa(k);
					}
					Curr_ms[key+i] = value;
					C_Links_loc it;
					it.file = filename;
					it.line = reader.linenum;
					it.ex = true;
					Curr_ms_loc[key+i] = it;
				}
			}
		}
	} catch(E_end_input) { }

	this->AddLink(linknew, Curr_ms, Curr_ms_loc);
}

void C_Links::AddLink(bool linknew, std::map<std::string,std::string>& Curr_ms, std::map<std::string,C_Links_loc>& Curr_ms_loc)
{
	if (!Curr_ms.empty())
	{
		if (linknew) {
			std::map<std::string,std::string>::const_iterator k;
			for(k = this->LinkDef.begin();k != this->LinkDef.end();k++) {
				Curr_ms.insert(*k);
			}
			std::string aka = Curr_ms["aka"];
			this->Vals[aka] = Curr_ms;
			this->Vals_loc[aka] = Curr_ms_loc;
			C_Links_loc str;
			str.file = this->_filename;
			str.line = this->_line;
			str.ex = false;
			this->Vals_defloc[aka] = str;
		} else {
			std::map<std::string,std::string>::const_iterator k;
			for(k = Curr_ms.begin();k != Curr_ms.end();k++) {
				this->LinkDef.insert(*k);
			}
		}
	}
	Curr_ms.clear();
}

C_LOG::C_LOG(std::string fname,std::string lvls,std::string br)
{
	this->is_stdout = strtolower(fname) == "stdout";
	this->lvls = lvls;
	unsigned int k;
	while ((k = br.find("\\r")) != std::string::npos) br.replace(k,2,"\r");
	while ((k = br.find("\\n")) != std::string::npos) br.replace(k,2,"\n");
	this->br = br;
	this->filename = fname;
	if ((!this->is_stdout) && (!(this->file = fopen(fname.c_str(),"a")))) {
		std::cout << "Open log file \"" << fname << "\" failed!";
		exit(1);
	}
	std::string Time = get_date("%c");
	this->write('1'," *** Started: " + Time + " (" + CONFIG.Constants["module"]+ "/" + VERSION + ")");
}

void C_LOG::write(char lvl, std::string message)
{
	if (this->lvls.find(lvl)!=std::string::npos) {
		std::string date = get_date("%d %b, %H:%M:%S");
		std::string msg = "";
		msg += lvl;
		msg += ' '+date+' '+message+' '+this->br;
		if (this->is_stdout) {
			std::cout << lvl << ' ' << date << ' ' << message << this->br;
		} else {
			fwrite(msg.c_str(),msg.length(),1,this->file);
		}
	}
}

C_LOG::~C_LOG()
{
	std::string Time = get_date("%c");
	this->write('1'," *** End: " + Time);
	if (!this->is_stdout) { 
		fclose(this->file); 
//		chmod($this->filename,0666);
	}
}

void C_LOG::refseek()
{
	if (!this->is_stdout)
		fseek(this->file,0,SEEK_END);
}

C_AreaDef::C_AreaDef(std::string line, bool group)
{
	int vkh = 0;
	std::string line2 = "";
	if (line != "") {
		int s = line.length();
		for (int i=0; i<s; i++) {
			char c = line[i];
			if (c == '"') {
	       		vkh++;
				if (vkh%2) {
					line2 += char(2);
				}
			} else if (line[i] == ' ' && (vkh%2)) {
				line2 += char(1);
			} else {
				line2 += c;
			}
		}
		std::list<std::string> msv;
		s = line2.length();
		int oldp = -1;
		for(int i=0;i<s;i++) {
			if (line2[i] == ' ') {
				if (i - oldp > 1) {
					msv.push_back(line2.substr(oldp+1,i-oldp-1));
				}
				oldp = i;
			}
		}
		if (s - oldp > 1) msv.push_back(line2.substr(oldp+1,std::string::npos));
		std::list<std::string>::iterator i;
		for (i=msv.begin();i!=msv.end();i++)
		{
			unsigned int k;
			while ((k = i->find(char(1))) != std::string::npos) i->replace(k,1," ");
			while ((k = i->find(char(2))) != std::string::npos) i->erase(k,1);
		}
		if (group) {
			if (!msv.empty()) {
				std::string frn = msv.front();
				if ((frn.length()<2) || (frn[0]!='-')) {
					this->GroupDescr = frn;
					msv.pop_front();
				}
			}
		} else {
			this->Status = strtolower(msv.front());
			msv.pop_front();
			this->EchoTag = msv.front();
			msv.pop_front();
			this->Path = msv.front();
			msv.pop_front();
		}
		while (!msv.empty())
		{
			std::string now = msv.front();
			if (now[0] == '-') {
				msv.pop_front();
				this->Options[now] = msv.front();
				msv.pop_front();
			} else {
				if (this->Linkattr.find(now) == this->Linkattr.end()) {
					this->Linkattr[now] = "";
				}
				msv.pop_front();
			}
		}
	}
}

void C_CONFIG::Open(std::string filename, std::string module)
{
	this->Constants["module"] = module;
	this->ReadCfgFile(filename);
	temp_dir = this->GetVar("tempdir");
}

void C_CONFIG::ReadCfgFile(std::string filename)
{
	C_configfile_reader reader(filename);
	try {
		while (1) {
			std::string line = reader.ReadStr();
//std::cout << line << "\n";
			if (line != "") {
				std::string key, value;
				split_2(' ',line,key,value);
				key = strtolower(key);
				if (key == "set") {
					std::string c,v;
					split_2('=',value,c,v);
					if (c != "" && (c[0] == '"') && (c[c.length()-1] == '"')) {
						c = c.substr(1,c.length()-2);
					}
					if (v != "" && (v[0] == '"') && (v[v.length()-1] == '"')) {
						v = v.substr(1,v.length()-2);
					}
					this->Constants[c] = v;
				} else if (key == "includelinks") {
					this->Links.OpenCfg(value);
				} else if (key == "includeareas") {
					this->Areas.OpenCfg(value);
				} else if (key == "includefareas") {
					this->FAreas.OpenCfg(value);
				} else if (key == "include") {
					this->ReadCfgFile(value);
				} else {
					std::string i = "";
					while (this->Vars.find(key+i) != this->Vars.end()) {
						int k = atoi(i.c_str())+1;
						i = my_itoa(k);
					}
					this->Vars[key+i] = value;
				}
			}
		}
	} catch(E_end_input) { }
}

C_configfile_reader::C_configfile_reader(std::string filename)
{
	if (filename != "") {
		this->nofile = 0;
		this->file = fopen(filename.c_str(),"r");
	} else {
		this->nofile = 1;
	}
	this->iflvl = 0;
	this->Levels.push(1);
	this->LvlsPs.push(1);
}

C_configfile_reader::~C_configfile_reader()
{
	if (!this->nofile) {
		fclose(this->file);
	}
}

std::string C_configfile_reader::ReadStr(std::string line)
{
	if (!this->nofile) {
		char buf[10240];
		char* result = fgets(buf,10240,this->file);
		if (result == NULL) {
			throw E_end_input();
		}
		line = result;
		this->linenum++;
	}
	line = trim(line);
	line = this->read_and_translate(line);

	if (line.substr(0,3) == "if ") {
		if (this->Levels.top()) {
			int eq = this->is_equal(line.substr(3,std::string::npos));
			this->LvlsPs.push(eq);
			this->Levels.push(eq);
		} else {
			this->Levels.pop();
			this->Levels.push(0);
			this->LvlsPs.pop();
			this->LvlsPs.push(1);
		}
		return "";
	} else if (line.substr(0,7) == "elseif ") {
		if (!this->LvlsPs.top()) {
			int eq = this->is_equal(line.substr(7,std::string::npos));
			this->LvlsPs.pop();
			this->Levels.pop();
			this->LvlsPs.push(eq);
			this->Levels.push(eq);
		} else {
			this->LvlsPs.pop();
			this->Levels.pop();
			this->LvlsPs.push(0);
			this->Levels.push(0);
		}
		return "";
	} else if (line.substr(0,line.find(' ')) == "endif") {
		this->iflvl--;
		this->Levels.pop();
		this->LvlsPs.pop();
		return "";
	} else if (this->Levels.top()) {
		return line;
	} else {
		return "";
	}
}

std::string C_configfile_reader::read_and_translate(std::string line)
{
	int ls = -1;
	int rs = -1;
	std::string l_chars = "";
	std::string l_spec = "";
	bool b, spec = false;
	int splen = 0;
	char c;

	int s = line.length();
	for (int i=0;i<s;i++) {
		c = line[i];
		if (c == '\\') {
			if (spec) {
				spec = false;
				l_chars += '\\';
				l_spec += char(0);
				splen++;
			} else {
				spec = true;
			}
		} else if ((b = (c == '[')) || (c == ']')) {
			if (spec) {
				l_chars += c;
				l_spec += char(0);
				splen++;
				spec = false;
			} else {
				if (b) {
					if (ls == -1) { ls = splen; }
				} else {
					if (rs == -1) { rs = splen; }
				}
				l_chars += char(0);
				l_spec += c;
				splen++;
			}
		} else if (c == '#') {
			if (spec) {
				l_chars += c;
				l_spec += char(0);
				splen++;
				spec = false;
			} else {
				break;
			}
		} else {
			if (spec) {
				l_chars += '\\';
				l_spec += char(0);
				splen++;
				spec = false;
			}
			l_chars += c;
			l_spec += char(1);
			splen++;
		}
	}
	if (spec) {
		l_chars += '\\';
		l_spec += char(0);
		splen++;
		spec = false;
	}
	s = splen;
	int bt = -1, et = -1;
	if (ls == -1) ls = std::string::npos;
	if (rs == -1) rs = std::string::npos;

	for (int i = 0; i < s; i++) {
		switch(l_spec[i]) {
			case('['):
				bt = i;
				et = -1;
				break;
			case(']'):
				et = i;
				if (bt!=-1) {
					this->insert_constant(l_spec, l_chars, bt, et);
					i = 0;
					s = l_spec.length();
					bt = -1;
				}
				break;
		}
	}

	std::string result = "";
	unsigned int o = 0;
	unsigned int l = l_chars.find(char(0));
	while (l != std::string::npos) {
		result += l_chars.substr(0,l) + l_spec[l];
		o = l;
		l = l_chars.find(char(0),l+1);
	}
	if (o == std::string::npos) o = 0;
	result += l_chars.substr(o,std::string::npos);
	return result;
}

int C_configfile_reader::is_equal(std::string str)
{
	unsigned int p = str.find("==");
	if (p==std::string::npos) return 1;
	std::string a = str.substr(0,p);
	std::string b = str.substr(p+2,std::string::npos);
	return (strtolower(a) == strtolower(b));
}

void C_configfile_reader::insert_constant(std::string &spec, std::string &chars, int b, int e)
{
	std::string cnst = chars.substr(b+1,e-b-1);
	std::string value = CONFIG.Constants[cnst];
	std::string spec_n = spec.substr(0,b);
	int s = value.length();
	for (int i=0;i<s;i++) spec_n += char(0);
	spec_n += spec.substr(e+1,std::string::npos);
	chars = chars.substr(0,b) + value + chars.substr(e+1,std::string::npos);
	spec = spec_n;
}

void C_Areas::OpenCfg(std::string filename)
{
	C_configfile_reader reader(filename);
	try {
		while (1) {
			std::string line = reader.ReadStr();
//std::cout << line << "\n";
			if (line != "") {
				if (strtolower(line.substr(0,line.find(' '))) == "echogroup") {
					std::string status, name, descr;
					split_2(' ',line,status,line);
					split_2(' ',line,name,descr);
					std::map<std::string,C_AreaDef*>::iterator iter;
					if ((iter = this->Groups.find(name)) != this->Groups.end()) delete (iter->second);
					this->Groups[name] = new C_AreaDef(descr,true);
				} else {
					this->AppendAreaToConf(line,filename,my_itoa(reader.linenum));
				}
			}
		}
	} catch(E_end_input) { }
}

std::string C_CONFIG::GetVar(std::string param)
{
	return this->Vars[param];
}

std::string C_CONFIG::GetLinkVar(std::string addr, std::string param)
{
	std::map<std::string, std::map<std::string,std::string> >::const_iterator iter;
	iter = this->Links.Vals.find(addr);
	if (iter == this->Links.Vals.end()) return "";
	std::map<std::string, std::string>::const_iterator link;
	link = iter->second.find(param);
	if (link == iter->second.end()) return "";
	return link->second;
}

void C_CONFIG::GetArray(std::string param, std::queue<std::string>& toarray)
{
	std::string i = "";
	std::map<std::string,std::string>::iterator iter;
	while ((iter = this->Vars.find(param+i)) != this->Vars.end()) {
		char buf[8];
		int k = atoi(i.c_str())+1;
		snprintf(buf,8,"%d",k);
		i = buf;
		std::string tmp = iter->second;
		toarray.push(tmp);
	}
}

std::string C_CONFIG::GetOurAkaFor(std::string link)
{
	std::string result = this->GetLinkVar(link,"ouraka");
	if (result == "") result = this->GetVar("address");
	return result;
}

void param_split(std::string line,std::queue<std::string>& toarray)
{
	bool kav = false;
	bool emp = false;
	std::string curr_add = "";
	int s = line.length();
	for(int i=0;i<s;i++) {
		switch(line[i]) {
			case('"'):
				if (kav) {
					kav = false;
					emp = false;
				} else {
					if (!emp) {
						toarray.push(curr_add);
						curr_add = "";
						emp = true;
					}
					kav = true;
				}
				break;
			case(' '):
				if (kav) {
					curr_add += ' ';
					emp = false;
				} else {
					if (!emp) {
						toarray.push(curr_add);
						curr_add = "";
						emp = false;
					}
				}
				break;
			default:
				curr_add += line[i];
				emp = false;
		}
	}
	toarray.push(curr_add);
}

void C_Areas::AppendAreaToConf(std::string line, std::string filename, std::string linenum)
{
	C_AreaDef* newad = new C_AreaDef(line);
	newad->l_file = filename;
	newad->l_line = linenum;
	std::map<std::string,std::string>::const_iterator i;
	if ((i = newad->Options.find("-g")) != newad->Options.end()) {
		std::string grp = i->first;
		std::map<std::string,C_AreaDef*>::const_iterator j;
		if ((j = this->Groups.find(grp)) != this->Groups.end()) {
			std::map<std::string,std::string>::const_iterator k;
			for(k=j->second->Options.begin();k!=j->second->Options.end();k++){
				newad->Options.insert(*k);
			}
			for(k=j->second->Linkattr.begin();k!=j->second->Linkattr.end();k++){
				newad->Options.insert(*k);
			}
// FIXME: Should be fixed for "^w" (aka not w) support and defaults saving.
		}
	}
	std::string etlo = strtolower(newad->EchoTag);
	std::map<std::string,C_AreaDef*>::iterator iter;
	if ((iter = this->Vals.find(etlo)) != this->Vals.end()) delete (iter->second);
	this->Vals[etlo] = newad;
}

C_Areas::~C_Areas()
{
	std::map<std::string,C_AreaDef*>::iterator iter;
	for (iter=CONFIG.Areas.Groups.begin(); iter!=CONFIG.Areas.Groups.end(); iter++)
		delete (iter->second);
	CONFIG.Areas.Groups.clear();
	for (iter=CONFIG.Areas.Vals.begin(); iter!=CONFIG.Areas.Vals.end(); iter++)
		delete (iter->second);
	CONFIG.Areas.Vals.clear();
}
