
/*\
  /     Amfow_imi - fido-over-email mailer (c) Alex Kocharin, 2:50/13
 /      
/       This program is free software, licence is GNU GPL
\*/


void initlog();
void donelog();
void tolog(char lvl,std::string message);
void refseeklog();
void param_split(std::string line,std::queue<std::string>& toarray);

std::string temp_dir;

struct C_Links_loc
{
	std::string file;
	int line;
	bool ex;
};

class C_Links
{
	private:
		std::string _filename;
		int _line;
		std::map<std::string,std::string> LinkDef;
		std::map<std::string,C_Links_loc> Vals_defloc;
	public:
		std::map<std::string, std::map<std::string,std::string> > Vals;
		std::map<std::string, std::map<std::string,C_Links_loc> > Vals_loc;
		void OpenCfg(std::string filename);
		void AddLink(bool linknew, std::map<std::string,std::string>& Curr_ms, std::map<std::string,C_Links_loc>& Curr_ms_loc);
};

class C_AreaDef
{
	public:
		std::string EchoTag;
		std::string Path;
		std::string Status;
		std::map<std::string,std::string> Linkattr;
		std::map<std::string,std::string> Options;
		std::string GroupDescr;
		std::string l_file;
		std::string l_line;
		C_AreaDef(std::string line, bool group = false);
};

class C_Areas
{
	public:
		void OpenCfg(std::string filename);
		void AppendAreaToConf(std::string line, std::string filename, std::string linenum);
		~C_Areas();
		std::map<std::string,C_AreaDef*> Vals;
		std::map<std::string,C_AreaDef*> Groups;
};

class C_CONFIG
{
	public:
		C_Links Links;
		C_Areas Areas;
		C_Areas FAreas;
		std::map<std::string,std::string> Constants;
		std::map<std::string,std::string> Vars;
		void Open(std::string filename, std::string module = "");
		void ReadCfgFile(std::string filename);
		std::string GetVar(std::string param);
		std::string GetLinkVar(std::string addr, std::string param);
		void GetArray(std::string param, std::queue<std::string>& toarray);
		std::string GetOurAkaFor(std::string link);
} CONFIG;

struct C_FidoAddr
{
	int zone;
	int net;
	int node;
	int point;
};

class Directory
{
	private:
		DIR *dir;
		struct dirent *file_st;
	public:
		bool is_opened() { return (dir != NULL); }
		Directory(std::string path);
		std::string read();
		~Directory();
};

class File
{
	private:
		FILE* fh;
	public:
		bool is_opened() { return (fh != NULL); }
		File(std::string file, std::string mode);
		bool Open(std::string file, std::string mode);
		void Close();
		File();
		~File();
		std::string Read(long size);
		void Write(std::string data);
		void WriteInt(short len, int data);
		int ReadInt(short len);
		std::string ReadAsciiZ();
		void WriteAsciiZ(std::string data);
		void WriteStr(std::string data, unsigned int len);
		std::string ReadStr(unsigned int len);
		void Seek(long ps, short mode = 0);
		std::string ReadS();
		char ReadC();
		bool Eof();
};

std::string my_itoa(int k)
{
	char buf[17];
	snprintf(buf,17,"%d",k);
	return buf;
}

C_FidoAddr GetFidoAddr(std::string addr, bool destroy = false)
{
	static regex_t fidoreg;
	static bool rxvalid;
	C_FidoAddr result;
	result.zone = 0;
	result.net = 0;
	result.node = 0;
	result.point = 0;
	if (destroy && rxvalid) {
		rxvalid = false;
		regfree(&fidoreg);
	} else if (addr != "") {
		if (!rxvalid) rxvalid = (regcomp(&fidoreg,"\\([0-9]\\+\\):\\([0-9]\\+\\)/\\([0-9]\\+\\)\\(\\.\\([0-9]\\+\\)\\)\\?",0) == 0);
		if (rxvalid) {
			regmatch_t matches[5];
			if (regexec(&fidoreg,addr.c_str(),6,&matches[0],0)==0) {
				result.zone = atoi(addr.substr(matches[1].rm_so,matches[1].rm_eo-matches[1].rm_so).c_str());
				result.net = atoi(addr.substr(matches[2].rm_so,matches[2].rm_eo-matches[2].rm_so).c_str());
				result.node = atoi(addr.substr(matches[3].rm_so,matches[3].rm_eo-matches[3].rm_so).c_str());
				if (matches[5].rm_so != -1)
					result.point = atoi(addr.substr(matches[5].rm_so,matches[5].rm_eo-matches[5].rm_so).c_str());
			}
		}
	}
	return result;
}

std::string JoinFidoAddr(C_FidoAddr addr)
{
	std::string result = my_itoa(addr.zone)+':'+my_itoa(addr.net)+'/'+my_itoa(addr.node);
	if (addr.point != 0) {
		result += '.';
		result += my_itoa(addr.point);
	}
	return result;
}

inline std::string basename(std::string file)
{
	unsigned int pos = file.rfind('/');
	if (pos == std::string::npos) { pos = 0; } else { pos++; };
	return file.substr(pos,std::string::npos);
}

std::string trim(std::string line)
{
	std::string fals = " \r\t\n";
	int l = -1;
	while (fals.find(line[++l])!=std::string::npos);
	int r = line.length();
	while (fals.find(line[--r])!=std::string::npos);
	return line.substr(l,r-l+1);
}

std::string chop(std::string line)
{
	std::string fals = "\r\n";
	int r = line.length();
	while (fals.find(line[--r])!=std::string::npos);
	return line.substr(0,r+1);
}

std::string strtolower(std::string str)
{
	for(unsigned int i = 0; i < str.length(); i++)
		str[i] = tolower(str[i]);
	return str;
}

std::string strtoupper(std::string str)
{
	for(unsigned int i = 0; i < str.length(); i++)
		str[i] = toupper(str[i]);
	return str;
}

std::string get_date(char* fmt, time_t t = (time_t)-1)
{
	char buf[80];
	if (t == (time_t)-1) t = time(0);
	strftime(buf,80,fmt,localtime(&t));
	return buf;
}

void split_2(char symbol, std::string line, std::string &result_key, std::string &result_value)
{
	std::string fals = " \r\t\n";
	unsigned int p = line.find(symbol);
	if (p == std::string::npos) {
		result_key = line;
		result_value = "";
	} else {
		int l = p;
		while (l && fals.find(line[--l]) != std::string::npos);
		int r = p;
		while (fals.find(line[++r]) != std::string::npos);
		result_key = line.substr(0,l+1);
		result_value = line.substr(r,std::string::npos);
	}
}

std::string rand_if_pkt(std::string fbase)
{
	std::string fbaselo = strtolower(fbase);
	int pos = fbase.find('.');
	bool is_pkt = false;
	if (pos == 8) {
		char ext1 = fbaselo[9];
		char ext2 = fbaselo[10];
		char ext3 = fbaselo[11];
		std::string allep = "odhic";
		if (allep.find(ext1) != std::string::npos) {
			if (ext2 == 'u' && ext3 == 't') {
				is_pkt = true;
				for(int k=0;k<8;k++) {
					int ord = (int)fbaselo[k];
					if (!((ord>=(int)'0' && ord<=(int)'9') || 
						(ord>=(int)'a' && ord<=(int)'f'))) {
						is_pkt = false;
						break;
					}
				}
			}
		}
	}
	std::string fas = fbase;
	if (is_pkt) {
		fas = my_itoa((10000000+((int)(((double)rand()/RAND_MAX)*89999999)))) + ".pkt";
	}
	return fas;
}

Directory::Directory(std::string path)
{
	this->dir = opendir(path.c_str());
}

Directory::~Directory()
{
	if (this->dir != NULL) closedir(this->dir);
}

std::string Directory::read()
{
	if (this->dir != NULL) {
		if ((this->file_st = readdir(this->dir)) != NULL) {
			return this->file_st->d_name;
		} else {
			return "";
		}
	} else {
		return "";
	}
}

std::string file_get_contents(std::string filename)
{
	long size;
	char* buffer;
	DIR* dir = opendir(filename.c_str());
	if (dir != NULL) {
		closedir(dir);
		return "";
	}
	FILE* fp = fopen(filename.c_str(), "r");
	if(fp == NULL)
		return "";
	fseek(fp, 0, SEEK_END);
	size = ftell(fp);
	rewind(fp);
	buffer = (char*)malloc(size+1);
	if(buffer == NULL)
		return "";
	fread(buffer,1,size,fp);
	fclose(fp);
	std::string ret = "";
	for(long i=0;i<size;i++) ret += buffer[i];
	free(buffer);
	return ret;
}

long filesize(std::string filename)
{
	long size;
	FILE* fp = fopen(filename.c_str(), "r");
	if(fp == NULL)
		return 0;
	fseek(fp, 0, SEEK_END);
	size = ftell(fp);
	fclose(fp);
	return size;
}

std::string base64_encode(unsigned char const* bytes_to_encode, unsigned int in_len);

std::string base64_encode(std::string s)
{
    return base64_encode(reinterpret_cast<const unsigned char*>(s.c_str()), s.length());
}

static const std::string base64_chars = 
             "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
             "abcdefghijklmnopqrstuvwxyz"
             "0123456789+/";


static inline bool is_base64(unsigned char c) {
  return (isalnum(c) || (c == '+') || (c == '/'));
}

std::string base64_encode(unsigned char const* bytes_to_encode, unsigned int in_len) {

  std::string ret;
  int i = 0;
  int j = 0;
  char char_array_3[3];
  char char_array_4[4];

  while (in_len--) {
    char_array_3[i++] = *(bytes_to_encode++);
    if (i == 3) {
      char_array_4[0] = (char_array_3[0] & 0xfc) >> 2;
      char_array_4[1] = ((char_array_3[0] & 0x03) << 4) + ((char_array_3[1] & 0xf0) >> 4);
      char_array_4[2] = ((char_array_3[1] & 0x0f) << 2) + ((char_array_3[2] & 0xc0) >> 6);
      char_array_4[3] = char_array_3[2] & 0x3f;

      for(i = 0; (i <4) ; i++)
        ret += base64_chars[char_array_4[i]];
      i = 0;
    }
  }

  if (i)
  {
    for(j = i; j < 3; j++)
      char_array_3[j] = '\0';

    char_array_4[0] = (char_array_3[0] & 0xfc) >> 2;
    char_array_4[1] = ((char_array_3[0] & 0x03) << 4) + ((char_array_3[1] & 0xf0) >> 4);
    char_array_4[2] = ((char_array_3[1] & 0x0f) << 2) + ((char_array_3[2] & 0xc0) >> 6);
    char_array_4[3] = char_array_3[2] & 0x3f;

    for (j = 0; (j < i + 1); j++)
      ret += base64_chars[char_array_4[j]];

    while((i++ < 3))
      ret += '=';

  }

  return ret;

}

std::string base64_decode(std::string const& encoded_string) {
  int in_len = encoded_string.size();
  int i = 0;
  int j = 0;
  int in_ = 0;
  unsigned char char_array_4[4], char_array_3[3];
  std::string ret;

  while (in_len-- && ( encoded_string[in_] != '=') && is_base64(encoded_string[in_])) {
    char_array_4[i++] = encoded_string[in_]; in_++;
    if (i ==4) {
      for (i = 0; i <4; i++)
        char_array_4[i] = base64_chars.find(char_array_4[i]);

      char_array_3[0] = (char_array_4[0] << 2) + ((char_array_4[1] & 0x30) >> 4);
      char_array_3[1] = ((char_array_4[1] & 0xf) << 4) + ((char_array_4[2] & 0x3c) >> 2);
      char_array_3[2] = ((char_array_4[2] & 0x3) << 6) + char_array_4[3];

      for (i = 0; (i < 3); i++)
        ret += char_array_3[i];
      i = 0;
    }
  }

  if (i) {
    for (j = i; j <4; j++)
      char_array_4[j] = 0;

    for (j = 0; j <4; j++)
      char_array_4[j] = base64_chars.find(char_array_4[j]);

    char_array_3[0] = (char_array_4[0] << 2) + ((char_array_4[1] & 0x30) >> 4);
    char_array_3[1] = ((char_array_4[1] & 0xf) << 4) + ((char_array_4[2] & 0x3c) >> 2);
    char_array_3[2] = ((char_array_4[2] & 0x3) << 6) + char_array_4[3];

    for (j = 0; (j < i - 1); j++) ret += char_array_3[j];
  }

  return ret;
}

File::File(std::string file, std::string mode)
{
	fh = fopen(file.c_str(),mode.c_str());
}

File::File()
{
	fh = NULL;
}

bool File::Open(std::string file, std::string mode)
{
	fh = fopen(file.c_str(),mode.c_str());
	return fh != NULL;
}

void File::Close()
{
	if (fh != NULL) {
		fclose(fh);
		fh = NULL;
	}
}

File::~File()
{
	if (fh != NULL) fclose(fh);
}

std::string File::Read(long size)
{
	if (fh == NULL) return "";
	char* buffer;
	buffer = (char*)malloc(size+1);
	if(buffer == NULL)
		return "";
	fread(buffer,1,size,fh);
	std::string ret = "";
	for(long i=0;i<size;i++) ret += buffer[i];
	free(buffer);
	return ret;
}

void File::Write(std::string data)
{
	if (fh == NULL) return;
	fwrite(data.c_str(), data.length(), 1, fh);
}

void File::WriteInt(short len, int data)
{
	if (fh == NULL) return;
	std::string result = "";
	for (short i=0; i<len; i++) {
		result += (char)((data >> (i*8)) & 255);
	}
	fwrite(result.c_str(), result.length(), 1, fh);
}

int File::ReadInt(short len)
{
	if (fh == NULL) return 0;
	int result = 0;
	for (short i=0; i<len; i++) {
		result += (fgetc(fh) << (i*8));
	}
	return result;
}

std::string File::ReadAsciiZ()
{
	if (fh == NULL) return "";
	std::string result = "";
	char c = fgetc(fh);
	while (c != 0) {
		result += c;
		c = fgetc(fh);
	}
	return result;
}

std::string File::ReadS()
{
	if (fh == NULL) return "";
	std::string result = "";
	if (feof(fh)) return "";
	char c = fgetc(fh);
	std::string fail = "\n";
	while ((!feof(fh)) && (fail.find(c)==std::string::npos)) {
		result += c;
		c = fgetc(fh);
	}
	return result + "\n";
}

char File::ReadC()
{
	if (fh == NULL || feof(fh)) return 0;
	return fgetc(fh);
}

void File::WriteAsciiZ(std::string data)
{
	if (fh == NULL) return;
	data += (char)0;
	fwrite(data.c_str(), data.length(), 1, fh);
}

void File::WriteStr(std::string data, unsigned int len)
{
	if (fh == NULL) return;
	if (data.length() > len) data = data.substr(0,len);
	std::string data2 = data;
	for (unsigned int i = data.length(); i<len; i++) {
		data2 += (char)0;
	}
	fwrite(data2.c_str(), data2.length(), 1, fh);
}

std::string File::ReadStr(unsigned int len)
{
	std::string res = Read(len);
	res = res.substr(0,res.find((char)0));
	return res;
}

void File::Seek(long ps, short mode)
{
	if (fh != NULL) fseek(fh,ps,mode);
}

bool File::Eof()
{
	if (fh == NULL) return false;
	return feof(fh);
}

bool fileexists(std::string file)
{
	FILE* f;
	f = fopen(file.c_str(),"r");
	bool result = f != NULL;
	if (result) fclose(f);
	return result;
}

std::string getcwd()
{
	char* buffer = (char*)malloc(1024);
	if(buffer == NULL)
		return "";
	getcwd(buffer,1023);
	std::string ret = buffer;
	free(buffer);
	return ret;
}
